import boto3
import traceback
import time
import os
import json
from .logger import get_logger

class Core:

    def __init__(self, debug):
        self.client = boto3.client('rds')
        self.logger = get_logger(debug)

    def get_cluster(self, cluster_identifier):
        clusters = self.get_clusters(cluster_identifier)
        return (None if len(clusters) == 0 else clusters[0])

    def get_clusters(self, cluster_identifier):
        option = {}
        if cluster_identifier: option['DBClusterIdentifier'] = cluster_identifier
        db_clusters = []
        try:
            db_clusters = self.client.describe_db_clusters(**option)['DBClusters']
        except Exception as e:
            return db_clusters

        return db_clusters

    def get_instance(self, identifier = None):
        instances = self.get_instances(identifier)
        return (None if len(instances) == 0 else instances[0])

    def get_instances(self, identifier):
        option = {}
        if identifier: option['DBInstanceIdentifier'] = identifier
        db_instances = []
        try:
            db_instances = self.client.describe_db_instances(**option)['DBInstances']
        except Exception as e:
            return db_instances

        return db_instances

    def wait_for_available(self, instance_identifier):
        waiter = self.client.get_waiter('db_instance_available')
        self.logger.info("waiting available instance... {instance_identifier}".format(instance_identifier=instance_identifier))
        waiter.wait(
            DBInstanceIdentifier=instance_identifier
        )

    def get_cluster_members(self, cluster_identifier):
        cluster = self.get_cluster(cluster_identifier)
        return (cluster["DBClusterMembers"] if cluster else [])

    def get_cluster_member_identifiers(self, cluster_identifier):
        members = self.get_cluster_members(cluster_identifier)
        return [member['DBInstanceIdentifier'] for member in members]

    def reboot_instance_and_wait(self, instance_identifier):
        response = self.client.reboot_db_instance(
            DBInstanceIdentifier=instance_identifier,
            ForceFailover=False
        )
        self.logger.info("rebooting instance... {instance_identifier}".format(instance_identifier=instance_identifier))
        self.wait_for_available(instance_identifier)

    def delete_instance_and_wait(self, instance_identifier):
        self.client.delete_db_instance(
            DBInstanceIdentifier=instance_identifier,
            SkipFinalSnapshot=True
        )
        self.logger.info("deleting instance... {instance_identifier}".format(instance_identifier=instance_identifier))
        waiter = self.client.get_waiter('db_instance_deleted')
        waiter.wait(
            DBInstanceIdentifier=instance_identifier
        )

    def delete_cluster_and_wait(self, cluster_identifier):
        cluster = self.get_cluster(cluster_identifier)
        if not cluster:
            self.logger.info("{cluster_identifier} is not exist.".format(cluster_identifier = cluster_identifier))
            return

        response = self.client.delete_db_cluster(
            DBClusterIdentifier=cluster_identifier,
            SkipFinalSnapshot=True
        )
        self.logger.info("deleting cluster... {cluster_identifier}".format(cluster_identifier=cluster_identifier))
        while len(self.get_clusters(cluster_identifier)) > 0:
            time.sleep(10)



import click
import sys
import os
import pprint
import time
from .logger import get_logger
from .core import *
from .util import *
from .validator import *
from datetime import datetime
from tabulate import tabulate

@click.group()
@click.option('--debug/--no-debug', default=False, help='enable debug logging')
def cli(debug):
    global core
    global client
    global logger

    core = Core(debug)
    client = core.client
    logger = get_logger(debug)

@cli.command(help = 'list instance and cluster')
def list():
    headers = [
        'DBInstanceIdentifier',
        'Engine',
        'DBInstanceStatus',
        'DBInstanceClass',
        'AvailabilityZone',
        'DBClusterIdentifier'
    ]
    rows = []
    for instance in core.get_instances(None):
        row = [instance.get(key) for key in headers]
        rows.append(row)
    print(tabulate(rows, headers = headers))

    print("")

    headers = [
        'DBClusterIdentifier',
        'Engine',
        'Status',
    ]
    rows = []
    for cluster in core.get_clusters(None):
        row = [cluster.get(key) for key in headers]
        row.append([m['DBInstanceIdentifier'] for m in cluster.get('DBClusterMembers')])
        rows.append(row)
    headers.append('DBClusterMembers')
    print(tabulate(rows, headers = headers))

@cli.command(help = 'list instance')
@click.option('--identifier', '-i', default=None)
def list_instance(identifier):
    db_instances = core.get_instances(identifier)
    Util.print_json(db_instances)

@cli.command(help = 'list clusters')
@click.option('--identifier', '-i', default=None)
def list_cluster(identifier):
    clusters = core.get_clusters(identifier)
    Util.print_json(clusters)

@cli.command(help = 'restore Aurora cluster and instance')
@click.option('--source_cluster_identifier', '-s', required=True)
@click.option('--target_cluster_identifier', '-t', required=True)
@click.option('--writer_instance_identifier', '-w', required=True)
@click.option('--reader_instance_identifier', '-r', default=[], multiple=True)
@click.option('--suffix', callback=Validator.validate_suffix, default=datetime.now().strftime('%Y%m%d%H%M%S'))
@click.pass_context
def restore(ctx, source_cluster_identifier, target_cluster_identifier, writer_instance_identifier, reader_instance_identifier, suffix):
    tmp_target_cluster_identifier = target_cluster_identifier + '-' + suffix
    tmp_writer_instance_identifier = writer_instance_identifier + '-' + suffix
    tmp_reader_instance_identifier =[i + '-' + suffix for i in reader_instance_identifier]
    identifiers = [tmp_writer_instance_identifier] + tmp_reader_instance_identifier

    source_cluster = core.get_cluster(source_cluster_identifier)
    writers = [member for member in source_cluster["DBClusterMembers"] if member['IsClusterWriter']]
    source_writer_instance = core.get_instance(writers[0]['DBInstanceIdentifier'])

    response = client.restore_db_cluster_to_point_in_time(
        SourceDBClusterIdentifier=source_cluster_identifier,
        DBClusterIdentifier=tmp_target_cluster_identifier,
        UseLatestRestorableTime=True,
        DBSubnetGroupName=source_cluster['DBSubnetGroup'],
        VpcSecurityGroupIds=[i['VpcSecurityGroupId'] for i in source_cluster['VpcSecurityGroups']],
        RestoreType='copy-on-write',
        Tags=[
            {
                'Key': 'Name',
                'Value': tmp_target_cluster_identifier
            },
        ]
    )

    time.sleep(3)

    create_db_instance_option = dict(
        DBInstanceClass=source_writer_instance['DBInstanceClass'],
        Engine='aurora',
        AvailabilityZone=source_writer_instance['AvailabilityZone'],
        DBSubnetGroupName=source_writer_instance['DBSubnetGroup']['DBSubnetGroupName'],
        DBParameterGroupName=source_writer_instance['DBParameterGroups'][0]['DBParameterGroupName'],
        DBClusterIdentifier=tmp_target_cluster_identifier
    )

    for identifier in identifiers:
        create_db_instance_option['DBInstanceIdentifier'] = identifier
        response = client.create_db_instance(**create_db_instance_option)

    ctx.invoke(delete_cluster, cluster_identifier=target_cluster_identifier)
    ctx.invoke(rename_tmp, cluster_identifier=tmp_target_cluster_identifier)

@cli.command(help = 'delete cluster and child instance')
@click.option('--cluster_identifier', '-i', required=True)
def rename_tmp(cluster_identifier):
    source_cluster = core.get_cluster(cluster_identifier)

    tmp_identifiers = [member['DBInstanceIdentifier'] for member in source_cluster["DBClusterMembers"]]

    # instance
    for tmp_identifier in tmp_identifiers:
        core.wait_for_available(tmp_identifier)
        tmp_identifier.rsplit('-', 1)
        identifier = tmp_identifier.replace('-' + tmp_identifier.rsplit('-', 1)[-1], '')
        logger.info("rename instance {0} => {1}".format(tmp_identifier, identifier))
        response = client.modify_db_instance(
            DBInstanceIdentifier=tmp_identifier,
            NewDBInstanceIdentifier=identifier,
            ApplyImmediately=True
        )

    # cluster
    new_cluster_identifier = cluster_identifier.replace('-' + cluster_identifier.rsplit('-', 1)[-1], '')
    logger.info("rename cluster {0} => {1}".format(cluster_identifier, new_cluster_identifier))
    response = client.modify_db_cluster(
        DBClusterIdentifier=cluster_identifier,
        NewDBClusterIdentifier=new_cluster_identifier,
        ApplyImmediately=True
    )

@cli.command(help = 'delete cluster and child instance')
@click.option('--cluster_identifier', '-i', required=True)
def delete_cluster(cluster_identifier):
    identifiers = core.get_cluster_member_identifiers(cluster_identifier)
    for identifier in identifiers:
        response = core.delete_instance_and_wait(identifier)

    core.delete_cluster_and_wait(cluster_identifier)

@cli.command(help = 'reboot cluster and child instance')
@click.option('--cluster_identifier', '-i', required=True)
def reboot_cluster(cluster_identifier):
    identifiers = core.get_cluster_member_identifiers(cluster_identifier)
    for identifier in identifiers:
        core.reboot_instance_and_wait(identifier)

@cli.command(help = 'run command')
@click.option('--command', '-c')
def run_command(command):
    return os.system(command)

def main():
    cli(obj={})



import logging
from logging import getLogger, StreamHandler, INFO, DEBUG

def get_logger(debug):
    log_fmt = '%(asctime)s- %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_fmt)

    logger = getLogger(__name__)
    if debug:
        logger.setLevel(DEBUG)
    else:
        logger.setLevel(INFO)
    return logger



import json
import datetime

class Util:

    @staticmethod
    def datetime_handler(x):
        if isinstance(x, datetime.datetime):
            return x.isoformat()
        raise TypeError("Unknown type")

    @staticmethod
    def print_json(data):
        print(
            json.dumps(
                data,
                indent=4,
                default=Util.datetime_handler
            )
        )



import click

class Validator:

    @staticmethod
    def validate_suffix(ctx, param, value):
        if '-' in value:
            raise click.BadParameter('Do not include "-"')
        else:
            return value



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open(os.path.join(here, 'README.md')).read()

version = '0.4.0'

install_requires = [
    'tabulate',
    'boto3',
    'click'
]

setup(name='auroraex',
    scripts=['bin/auroraex'],
    version=version,
    description="Command Line utility for Amazon Aurora.",
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    keywords='amazon aurora aws mysql',
    author='Hiroshi Toyama',
    author_email='toyama0919@gmail.com',
    url='https://github.com/toyama0919/auroraex',
    license='MIT',
    packages=find_packages('src'),
    package_dir = {'': 'src'},include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={
        'console_scripts':
            ['auroraex=auroraex.commands:main']
    }
)
